#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "data.h"

int mlit0_alloc_cnt;
int mcl0_alloc_cnt;
int mlit1_alloc_cnt;
int mcl1_alloc_cnt;
int modwff_alloc_cnt;
int mlit0_free_cnt;
int mcl0_free_cnt;
int mlit1_free_cnt;
int mcl1_free_cnt;
int modwff_free_cnt;
int propwff_free_cnt;
int propwff_alloc_cnt;
int propcl_free_cnt;
int propcl_alloc_cnt;


/***********************************************************************
***********************************************************************
           MODAL WFFS STRUCTURES
***********************************************************************
***********************************************************************

MODWFF: (MODAL WFF)
+----------+
|nvars     |
|dim       |     _0___________________________dim-1
|modcls----+---->| |*| | | | | | | | | | | | | | 
+----------+     ---|---------------------------
                    |
                    v
                   MODCL

MODCL:  (MODAL CLAUSE)
+----------+
|          |     _0___2_
|modlits---+---->| |*| |
+----------+     ---|---
                    |
                    v
                   MODLIT

MODLIT:  (MODAL LITERAL)
+-------+
|code   |    
|modcl--+---->MODCL
+-------+     

CODE: 
=   ISMODAL   => modal literal, positive "(box clause)"
=   -ISMODAL  => modal literal, negative "(not box clause)"
=/= ISMODAL   => propositional literal, positive or negative 
                 A_n o (not A_n) [modcl becomes unrelevant]

******************************************************/


/***********************************************************************
      PRIMITIVES for MODAL WFFS STRUCTURES
***********************************************************************
<obj1>-set-<obj2> : sets the value of <obj2> inside <obj1>
<obj1>-get-<obj2> : gets the value of <obj2> from <obj1>
<property>_<obj>  : does <obj> verify <property>?
***********************************************************************/

void modwff_set_dim(modwff *w_p,int dim)
{
 w_p->dim=dim;
}
 
void modwff_set_nvars(modwff *w_p,int nvars)
{
 w_p->nvars=nvars;
}
 
void modwff_set_modcls(modwff *w_p,modcl **modcls)
{
 w_p->modcls=modcls;
}

void modwff_set_modcl(modwff *w_p,int i,modcl *modcl)
{
 w_p->modcls[i]=modcl;
}

modcl  **modwff_get_modcls(modwff *w_p)
{
  return w_p->modcls;
}

modcl  *modwff_get_modcl(modwff *w_p,int i)
{
  return w_p->modcls[i];
}

int modwff_get_dim(modwff *w_p)
{
  return w_p->dim;
}

int modwff_get_nvars(modwff *w_p)
{
  return w_p->nvars;
}


/******************************************************
                 modal clauses
******************************************************/

void modcl_set_modlits(modcl *c_p,modlit *modlits[])
{
 c_p->modlits=modlits;
}

void modcl_set_modlit(modcl *c_p,int i,modlit *modlit)
{
 c_p->modlits[i]=modlit;
}

modlit  **modcl_get_modlits(modcl *c_p)
{
  return c_p->modlits;
}

modlit  *modcl_get_modlit(modcl *c_p,int i)
{
  return c_p->modlits[i];
}

int modcl_get_dim (modcl *c_p) {return 3;}

/******************************************************
                   modal literals
******************************************************/
void modlit_set_modcl(modlit *l_p,modcl *c)
{
  l_p->modcl=c;
}

void modlit_set_code(modlit *l_p,int code)
{
 l_p->code=code;
}

modcl *modlit_get_modcl(modlit *l_p)
{
  return l_p->modcl;
}

int modlit_get_code(modlit *l_p)
{
  return l_p->code;
}

int pos_modlit(modlit *l_p)
{
  return (l_p->code>0);
}

int neg_modlit(modlit *l_p)
{
  return (l_p->code<0);
}

int modal_modlit(modlit *l_p)
{
  return (l_p->code == ISMODAL)||(l_p->code == -ISMODAL);
}

int prop_modlit(modlit *l_p)
{
  return !((l_p->code == ISMODAL)||(l_p->code == -ISMODAL));
}




/****************************************************************************/
/*  SPECIAL PURPOSE                                                         */
/****************************************************************************/
int mlit_same(modlit * l_p1,modlit * l_p2,int depth); 
int mcl_same(modcl *mcl1,modcl *mcl2,int depth);
int mlit_sameatom(modlit *l_p1,modlit *l_p2,int depth); 


int mlit_same(modlit * l_p1,modlit * l_p2,int depth) {
//printf("mlit_same, depth %d\n",depth);
  if ((l_p1->code)!=(l_p2->code))
     return 0;
  if (depth) 
     return mcl_same(modlit_get_modcl(l_p1),modlit_get_modcl(l_p2),depth-1);
  return 1;
}


int mcl_same(modcl *mcl1,modcl *mcl2,int depth)
{ 
  int i;
//printf("mcl_same, depth %d\n",depth);
//printf(" mcl1: ");print_modcl(mcl1,depth);
//printf(" mcl2: ");print_modcl(mcl2,depth);
  for(i=0;i<3;i++) 
    if (!mlit_same(modcl_get_modlit(mcl1,i),modcl_get_modlit(mcl2,i),depth))
       return 0;
  return 1;
}
  
int mlit_sameatom(modlit *l_p1,modlit *l_p2,int depth) {
  modcl *mcl1;modcl *mcl2;
//printf("mlit_sameatom, depth %d, ",depth);
//if (!l_p1) printf("l_p null "); else print_modlit(l_p1,depth);
//if (!l_p2) printf("l_p null "); else print_modlit(l_p2,depth);
  if ((l_p1==NULL) || (l_p2==NULL)) {
//printf("  mlit_sameatom returns 0\n");

    return 0;
}
//printf("\n");
  mcl1=modlit_get_modcl(l_p1);
  mcl2=modlit_get_modcl(l_p2);
  return mcl_same(mcl1,mcl2,depth-1);
}   
int mlit1_sameatom(modlit *l_p1,modlit *l_p2) {
   return mlit_sameatom(l_p1,l_p2,1);
}
  

/******************************************************

allocation primitives for modal wffs

******************************************************/

/******************************************************
 allocates room for a modal literal
******************************************************/
modlit *mlit_alloc(int depth);
modcl * mcl_alloc(int depth);
modwff *modwff_alloc(int dim,int depth);

/******************************************************
 allocates room for a modcl
******************************************************/
modcl * mcl_alloc(int depth) {
  int i;
  modcl *cl_p;
  cl_p = (modcl *) calloc(1,sizeof(modcl));
  cl_p->modlits=(modlit **) calloc(3,sizeof(modlit*));
  for(i=0;i<3;i++) 
     cl_p->modlits[i] = mlit_alloc(depth);
  return cl_p;
}

/******************************************************
 allocates room for a modlit
******************************************************/
modlit *mlit_alloc(int depth)
{ 
  
  modlit *lit_p;
  lit_p = (modlit *) calloc(1,sizeof(modlit));
  if (depth==0)
     lit_p->modcl = NULL_CLAUSE;
  else      
     lit_p->modcl = mcl_alloc(depth-1);     
  return lit_p;
}

/******************************************************
 allocates room for a modwff
******************************************************/
modwff *modwff_alloc(int dim,int depth)
  {modwff *wff_p;
   int i;
   wff_p = (modwff *) calloc(1,sizeof(modwff));
   wff_p->dim = dim;
   wff_p->modcls = (modcl **) calloc(dim,sizeof(modcl *));
   for (i=0;i<dim;i++)
     wff_p->modcls[i]= mcl_alloc(depth);
  return wff_p;
  }



/******************************************************
*******************************************************

         PRINT FUNCTIONS

     print modal literals, clauses, wffs

******************************************************/
void print_modcl(FILE *outfile, modcl *c_p,int grade);

/******************************************************
 prints a modal literal
******************************************************/
void print_modlit(FILE *outfile, modlit *l_p,int grade)
{
 if (prop_modlit(l_p)) {
   fprintf(outfile, "%d ",modlit_get_code(l_p));
 }
 else {
   if (neg_modlit(l_p)) 
     fprintf(outfile, "-");
   /***Modified by Armando Tacchella, January 1998 
     fixed modality to support the new input language of KsatC */
   fprintf(outfile, "#(1)(");
   print_modcl(outfile, modlit_get_modcl(l_p),grade-1); 
   fprintf(outfile, ")");
 }
 if (grade) 
   fprintf(outfile, " "); 
}
 
/******************************************************
 prints a modal clause
******************************************************/
void print_modcl(FILE *outfile, modcl *c_p,int grade)
{int i;
 fprintf(outfile, (grade? " v(" : "v("));
 for (i=0;i<modcl_get_dim(c_p);i++)
   print_modlit(outfile, modcl_get_modlit(c_p,i),grade);
 fprintf(outfile, (grade? ")" : ")"));
}
 
/******************************************************
 prints a modal wff
******************************************************/
void print_modwff(FILE *outfile, modwff *w_p)
{
 int i;
 fprintf(outfile, "wff = \n");
 fprintf(outfile, "^(\n");
 for (i=0;i<modwff_get_dim(w_p);i++) {
    print_modcl(outfile, modwff_get_modcl(w_p,i),1);
    fprintf(outfile, "\n"); 
 }
 fprintf(outfile, ")\n");
} 


/******************************************************
*******************************************************

         PRINT FUNCTIONS FOR PROLOG

******************************************************/
void prologprint_modcl(FILE *outfile, modcl *c_p,int grade, int flag_comma);

/******************************************************
 prints a modal literal
******************************************************/
void prologprint_modlit(FILE *outfile, modlit *l_p,int grade, int flag_comma)
{
 if (prop_modlit(l_p)) {
   if (modlit_get_code(l_p) < 0)
     fprintf(outfile, (flag_comma ? "not(c%d)" : "not(c%d),"),abs(modlit_get_code(l_p)));
   else
     fprintf(outfile, (flag_comma ? "c%d" : "c%d,"),modlit_get_code(l_p));
 }
 else {
   if (neg_modlit(l_p)) {
     fprintf(outfile, "not(all(r1,");
     prologprint_modcl(outfile, modlit_get_modcl(l_p),grade-1,1); 
     fprintf(outfile, flag_comma ? "))" : ")),");
   } else {
     fprintf(outfile, "all(r1,");
     prologprint_modcl(outfile, modlit_get_modcl(l_p),grade-1,1); 
     fprintf(outfile, flag_comma ? ")" : "),");
   }
 }
 /* if (grade)  */
/*    fprintf(outfile, " ");  */
}
 
/******************************************************
 prints a modal clause
******************************************************/
void prologprint_modcl(FILE *outfile, modcl *c_p,int grade, int flag_comma)
{int i;
 fprintf(outfile, "or([");
 for (i=0;i<modcl_get_dim(c_p);i++)
   prologprint_modlit(outfile, modcl_get_modlit(c_p,i),grade,
                (i == (modcl_get_dim(c_p) - 1)));
 fprintf(outfile, flag_comma ? "])" : (grade ? "]),\n" : "]),"));
}
 
/******************************************************
 prints a modal wff
******************************************************/
void prologprint_modwff(FILE *outfile, modwff *w_p)
{
 int i;
 fprintf(outfile, "and([\n");
 for (i=0;i<modwff_get_dim(w_p);i++) {
    prologprint_modcl(outfile, modwff_get_modcl(w_p,i),1, 
                (i == (modwff_get_dim(w_p) - 1)));
    fprintf(outfile, "\n");  
 }
 fprintf(outfile, "]).\n");

} 

/******************************************************
*******************************************************

         PRINT FUNCTIONS FOR LISP

******************************************************/

void lispprint_modcl(FILE *outfile,modcl *c_p,int grade);

/******************************************************
 prints a modal literal
******************************************************/

void lispprint_modlit(FILE *outfile,modlit *l_p,int grade)
{
 int lit;
 if (prop_modlit(l_p)) {
   lit=modlit_get_code(l_p);
   if (lit>0)
      fprintf(outfile," A%d",lit-1);
   else 
      fprintf(outfile," (not A%d)",(lit<0? -lit :lit)-1);
 }
 else {
   if (neg_modlit(l_p)) 
     fprintf(outfile," (not");
   fprintf(outfile," (box");
   lispprint_modcl(outfile,modlit_get_modcl(l_p),grade-1); 
   fprintf(outfile,")");
   if (neg_modlit(l_p))
     fprintf(outfile,")");
 }
}
 
/******************************************************
 prints a modal clause
******************************************************/
void lispprint_modcl(FILE *outfile,modcl *c_p,int grade)
{
 int i,dim=modcl_get_dim(c_p);
 if (dim!=0){
  if (dim ==1){
    lispprint_modlit(outfile,modcl_get_modlit(c_p,0),grade);
  }
  else {
    for (i=0;i<dim-1;i++){
      fprintf(outfile," (or");
      lispprint_modlit(outfile,modcl_get_modlit(c_p,i),grade);
    }
    lispprint_modlit(outfile,modcl_get_modlit(c_p,dim-1),grade);
    for (i=0;i<dim-1;i++)
      fprintf(outfile,")");
}}}
 
/******************************************************
 prints a modal wff
******************************************************/
void lispprint_modwff(FILE *outfile,modwff *w_p)
{
 int i,dim=modwff_get_dim(w_p);
 if (dim!=0){
  if (dim ==1){
    lispprint_modcl(outfile,modwff_get_modcl(w_p,0),1);
    fprintf(outfile,"\n");
  }
  else {
    for (i=0;i<dim-1;i++){
      fprintf(outfile," (and");
      lispprint_modcl(outfile,modwff_get_modcl(w_p,i),1);
      fprintf(outfile,"\n");
    }
    fprintf(outfile,"     ");
    lispprint_modcl(outfile,modwff_get_modcl(w_p,dim-1),1);
    fprintf(outfile,"\n");
    for (i=0;i<dim-1;i++)
      fprintf(outfile,")");
    fprintf(outfile,"\n");
}}}


/******************************************************
*******************************************************

       PRINT FUNCTIONS FOR KRIS (concepts)

******************************************************/
void krisprint_modcl(FILE *outfile, modcl *c_p,int grade, int flag_comma);

/******************************************************
 prints a modal literal
******************************************************/
void krisprint_modlit(FILE *outfile, modlit *l_p,int grade, int flag_comma)
{
 if (prop_modlit(l_p)) {
   if (modlit_get_code(l_p) < 0)
     fprintf(outfile, (flag_comma ? 
                       "(NOT C%d)" : 
                       "(NOT C%d) "),abs(modlit_get_code(l_p)));
   else
     fprintf(outfile, (flag_comma ? 
                       "C%d" : 
                       "C%d "),modlit_get_code(l_p));
 }
 else {
   if (neg_modlit(l_p)) {
     fprintf(outfile, "(NOT (ALL R1 ");
     krisprint_modcl(outfile, modlit_get_modcl(l_p),grade-1,1); 
     fprintf(outfile, flag_comma ? ")) " : "))");
   } else {
     fprintf(outfile, "(ALL R1 ");
     krisprint_modcl(outfile, modlit_get_modcl(l_p),grade-1,1); 
     fprintf(outfile, flag_comma ? ") " : ")");
   }
 }
 /* if (grade)  */
/*    fprintf(outfile, " ");  */
}
 
/******************************************************
 prints a modal clause
******************************************************/
void krisprint_modcl(FILE *outfile, modcl *c_p,int grade, int flag_comma)
{int i;
 fprintf(outfile, "(OR ");
 for (i=0;i<modcl_get_dim(c_p);i++)
   krisprint_modlit(outfile, modcl_get_modlit(c_p,i),grade,
                (i == (modcl_get_dim(c_p) - 1)));
 fprintf(outfile, flag_comma ? ") " : (grade ? ")\n" : ")"));
}
 
/******************************************************
 prints a modal wff
******************************************************/
void krisprint_modwff(FILE *outfile, modwff *w_p)
{
 int i;
 fprintf(outfile, "(AND ");
 for (i=0;i<modwff_get_dim(w_p);i++) {
    krisprint_modcl(outfile, modwff_get_modcl(w_p,i),1, 
                (i == (modwff_get_dim(w_p) - 1)));
 }
 fprintf(outfile, ")\n");

} 
