#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <math.h>
#include <limits.h> 

#define ISMODAL 10000 
#define OK 0
#define ERR -1        
#define NULL_CLAUSE NULL
#define NIL -1
#define MAX_CL_LEN 3

#define FIRSTRED 10
#define LASTRED  15
//#define ENHANCED

#define int_abs(a) (a < 0)? -a : a 
#define int_sign(a) (a < 0)? -1 : 1 
#define odd(a)    (a%2)
#define myodd(a)    (odd(a)? -1 : 1)
#define mynegmak(a) (odd(a)? -a : a)
#define min(a,b) (a<b ? a : b)
#define max(a,b) (a>b ? a : b)
#define valcomp(a,b) (a<b ? -1 : (a>b ? 1 : 0))



/***********************************************************************
***********************************************************************
           MODAL WFFS STRUCTURES
***********************************************************************
***********************************************************************

MODWFF: (MODAL WFF)
+----------+
|nvars     |
|dim       |     _0___________________________dim-1
|modcls----+---->| |*| | | | | | | | | | | | | | 
+----------+     ---|---------------------------
                    |
                    v
                   MODCL


MODCL:  (MODAL CLAUSE)
+----------+
|          |     _0_1_2_
|modlits---+---->| |*| | 
+----------+     ---|---
                    |
                    v
                   MODLIT


MODLIT:  (MODAL LITERAL)
+-------+
|code   |    
|modcl--+---->MODCL
+-------+     

CODE: 
=   ISMODAL   => modal literal, positive "(box clause)"
=   -ISMODAL  => modal literal, negative "(not box clause)"
=/= ISMODAL   => propositional literal, positive or negative 
                 A_n o (not A_n) [modcl becomes unrelevant]

******************************************************/

typedef struct mwff { 
 int          nvars;        /* number of variables: A_1...A_{nvars} */        
 int          dim;          /* number of clauses */
 struct mcl   **modcls;     /* array[1..dim] of pointers to modcls*/ 
} modwff;                   /* MODAL WFF */


typedef struct  mcl {
 int          dim ;         /* number of literals (def 3) */
 struct mlit  **modlits ;   /* array[1..dim] of pointers to modlits */
} modcl;                    /* MODAL CLAUSE */


typedef struct mlit {         
 int          code;         
 struct mcl   *modcl;               
} modlit;                   /* MODAL LITERAL */


/***********************************************************************
      PRIMITIVES for MODAL WFFS STRUCTURES
***********************************************************************
<obj1>-set-<obj2> : sets the value of <obj2> inside <obj1>
<obj1>-get-<obj2> : gets the value of <obj2> from <obj1>
<property>_<obj>  : does <obj> verify <property>?
***********************************************************************/

void modwff_set_dim(modwff *w_p,int dim);
void modwff_set_nvars(modwff *w_p,int nvars);
void modwff_set_modcls(modwff *w_p,modcl **modcls);
void modwff_set_modcl(modwff *w_p,int i,modcl *modcl);
void modcl_set_dim(modcl *c_p,int dim);
void modcl_set_modlits(modcl *c_p,modlit *modlits[]);
void modcl_set_modlit(modcl *c_p,int i,modlit *modlit);
void modlit_set_modcl(modlit *l_p,modcl *c);
void modlit_set_code(modlit *l_p,int code);

modcl  **modwff_get_modcls(modwff *w_p);
modcl  *modwff_get_modcl(modwff *w_p,int i);
int modwff_get_dim(modwff *w_p);
int modwff_get_nvars(modwff *w_p);
modlit  **modcl_get_modlits(modcl *c_p);
modlit  *modcl_get_modlit(modcl *c_p,int i);
int modcl_get_dim(modcl *c_p);
modcl *modlit_get_modcl(modlit *l_p);
int modlit_get_code(modlit *l_p);


int pos_modlit(modlit *l_p);
int neg_modlit(modlit *l_p);
int modal_modlit(modlit *l_p);
int prop_modlit(modlit *l_p);

int mcl0_same(modcl *mcl1,modcl *mcl2);
int mlit1_sameatom(modlit *l_p1,modlit *l_p2);

modlit *mlit_alloc(int depth);
modcl * mcl_alloc(int depth);
modwff *modwff_alloc(int dim,int depth);

/* void print_modcl(modcl *c_p,int grade);; */
/* void print_modlit(modlit *l_p,int grade); */
/* void print_modcl(modcl *c_p,int grade); */
void print_modwff(FILE *outfile, modwff *w_p);

/* PROLOG PRINT ROUTINE */

void prologprint_modwff(FILE *outfile, modwff *w_p);

/* LISP PRINT ROUTINE */

void lispprint_modwff(FILE *outfile, modwff *w_p);

/* KRIS PRINT ROUTINE */

void krisprint_modwff(FILE *outfile, modwff *w_p);



