//****************************************************************************
//                                                                            
//  The command:                                                              
//                                                                            
//  > generate <clause #> <var #> <propositional > <depth> [seed]             
//                                                                            
//  generates random modal 3CNF wffs according to the following parameters:    
//                                                                            
//  <clause #>        : number of clauses                                     
//  <var #>           : number of propositional variable                      
//  <propositional %> : percentage of propositional atoms (suggested: 0)      
//  <depth>           : modal depth                                           
//  [seed]            : seed (optional). if -1, random. if omitted, 0.  
//                                                                            
//  The syntax is the following:                                              
//                                                                            
//  <wff>      ::= ^( [<clause>]* )                                           
//  <clause>   ::= v( <literal> <literal> <literal> )                         
//  <literal>  ::= [-]<atom>                                                  
//  <atom>     ::= <propatom> | <modatom>                                     
//  <propatom> ::= <integer>                                                  
//  <modatom>  ::= #<clause>                                                  
//                                                                            
//****************************************************************************

//****************************************************************************
//                                                                            
//  SETTINGS :                                                                 
//                                                                            
//  if MULTISET, repetitions in clauses are allowed. EX: v(1 -3 1), v(1 -1 4)  
//  if SORTED, clauses are generated sorted                                   
//  if SORT_ABSSIGN, literals are ordered following the priorities:        
//       1) propositional < modal                                             
//       2) n-1 < n                                                           
//       3) -lit < lit                                                        
//  otherwise, literals are ordered following the priorities:              
//       1) -lit < lit                                                        
//       2) propositional < modal                                             
//       3) n-1 < n                                                           
//        
// NOTE: for the hardest formulas, set 
// //#define MULTISET
// #define SORTED
//        
//****************************************************************************

//#define MULTISET
#define SORTED
#define MULTIMOD
#define SORT_ABSSIGN


/****************************************************************************/
/*                                                                          */
/*                 RANDOM MACROS                                        */
/*                                                                          */
/****************************************************************************/

#define RANDOM_MAX ((1<<30) -1 + (1<<30))   /* 2^31-1 */ 

/* long random(void); */
/* void srandom(unsigned int); */

/*--------------------------------------------------------------------------*/
/*  initializes random *really* randomly                                    */
/*--------------------------------------------------------------------------*/  
#define time_rand_init() (srandom(time(NULL)))

/*--------------------------------------------------------------------------*/
/*  initializes random with a given seed  (unsigned long seed)              */
/*--------------------------------------------------------------------------*/
#define rand_init(seed) (srandom(seed))

/*--------------------------------------------------------------------------*/
/*  Gets a random number between 0 and x-1                                  */
/*--------------------------------------------------------------------------*/  
#define rand_get(max) (random()%max)

/* randsign returns 1 with probability prob, 0 with probability 1-prob */
/* randprob returns 1 with probability prob, -1 with probability 1-prob */
/* randint returns a random integer in [1..max] */
/* NOTE: randint is reasonably uniform only for max << RANDOM_MAX */
#define randsign(prob) ((random() < (RANDOM_MAX)*prob)? 1 : -1)
#define randprob(prob) (random() < (RANDOM_MAX)*prob) 
#define randint(max) (random()%max +1)

static void rand_mlit_fill(modlit *l_p,int varnum,float prop_prob,int depth);
static void rand_mcl_fill(modcl *c_p,int varnum,float prop_prob,int depth);
void rand_modwff_fill(modwff *w_p,int varnum,float prop_prob,int depth);


void swap_lits(modlit *l1,modlit *l2);
void sort_modcl(modcl *mc);
int code_min(int c1,int c2);
int modlit_min(modlit *l1,modlit *l2);
int modcl_min(modcl *c1,modcl *c2);

int modlit_equal(modlit *l1,modlit *l2);
int modcl_equal(modcl *c1,modcl *c2);

