#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "random_set.h"


/****************************************************************************/
/*                                                                          */
/*         RANDOM SETS                                                      */
/*                                                                          */
/****************************************************************************/

/*----------------------------------------------------------------*/
/* Given N and num, it implements a BIJECTIVE function            */
/* which maps a number in 0..N*(N-1)*(N-2)-1 into an unsorted     */
/* set (a1 a2 a3), where ai in 1..N and ai =/= aj for every i,j   */
/*----------------------------------------------------------------*/
   void num2array3 (int num, int N, int A[3]) 
{
  int a0,a1,a2;
  a0 = num%N;
  num  /= N;
  a1 = num%(N-1);
  a2 = num/(N-1);

  a2 += (a2>=a1);
  a2 += (a2>=a0); 
  a1 += (a1>=a0);
  A[0]= ++a0;A[1]= ++a1;A[2]= ++a2;
}

/*----------------------------------------------------------------*/
/* Given N and num, it implements a BIJECTIVE function            */
/* which maps a number in 0..N*(N-1)-1 into an unsorted           */
/* set (a1 a2), where ai in 1..N and ai =/= aj for every i,j      */
/*----------------------------------------------------------------*/
   void num2array2 (int num, int N, int A[2]) 
{
  int a0,a1;
  a0 = num%N;
  a1 = num/N;
  a1 += (a1>=a0);
  A[0]= ++a0;A[1]= ++a1;
}

/*--------------------------------------------------------------------------*/
/*  Given an array of 3 integers and a code in 0..7                         */
/*  it negates some of them according to the bit pattern                    */
/*  suggested by code                                                       */
/*--------------------------------------------------------------------------*/
void negate_array3(int code,int A[3])
{
  if (code%2) A[2] *= -1; 
  code /= 2;
  if (code%2) A[1] *= -1;
  if (code/2) A[0] *= -1;
}

/*----------------------------------------------------------------*/
/* Sorts an array of 3 elements                                   */ 
/*----------------------------------------------------------------*/
   void sort_array3(int A[3])
{ int tmp;
  if (A[2]<A[1]) {tmp=A[2];A[2]=A[1];A[1]=tmp;}
  if (A[1]<A[0]) {tmp=A[1];A[1]=A[0];A[0]=tmp;}
  if (A[2]<A[1]) {tmp=A[2];A[2]=A[1];A[1]=tmp;}
}

/* ---------------------------------------------------------------- 
int main() 
{
   int num,N,nmax;
   int A[3];
      
   printf("N=? "); scanf("%d",&N);
   nmax=N*(N-1)*(N-2);

      printf("\n");
   for (num=0;num<nmax;num++)
     {
      num2array3(num,N,A);
      printf("num= %d, (%3d %3d %3d)\n",
      num,A[2],A[1],A[0]
      );
     }
      printf("\n");

}
 ----------------------------------------------------------------*/



